"""A module controlling the app."""
import logging
from dataclasses import fields, is_dataclass
from typing import Any, Dict, Optional, Type, Union, cast

from epplib.commands.base import Request, Result
from epplib.commands.check import CheckContact, CheckDomain, CheckKeyset, CheckNsset
from epplib.commands.command_extensions import (CreateContactMailingAddressExtension, CreateDomainEnumExtension,
                                                RenewDomainEnumExtension, UpdateContactMailingAddressExtension,
                                                UpdateDomainEnumExtension)
from epplib.commands.create import CreateContact, CreateDomain, CreateKeyset, CreateNsset
from epplib.commands.delete import DeleteContact, DeleteDomain, DeleteKeyset, DeleteNsset
from epplib.commands.extensions import (CreditInfoRequest, SendAuthInfoContact, SendAuthInfoDomain, SendAuthInfoKeyset,
                                        SendAuthInfoNsset, TestNsset)
from epplib.commands.info import InfoContact, InfoDomain, InfoKeyset, InfoNsset
from epplib.commands.list import (GetResults, ListContacts, ListDomains, ListDomainsByContact, ListDomainsByKeyset,
                                  ListDomainsByNsset, ListKeysets, ListKeysetsByContact, ListNssets,
                                  ListNssetsByContact, ListNssetsByNs)
from epplib.commands.renew import RenewDomain
from epplib.commands.transfer import TransferContact, TransferDomain, TransferKeyset, TransferNsset
from epplib.commands.update import UpdateContact, UpdateDomain, UpdateKeyset, UpdateNsset

from .commands.argument_registry import ArgumentRegistry, register_defaults
from .commands.command import Command
from .commands.command_registry import CommandRegistry
from .commands.help import HelpCommand
from .commands.service import ServiceCommand, ServiceExit, ServiceLogin, ServiceLogout, ServiceSwitch
from .settings import Settings

LOGGER = logging.getLogger(__name__)


class Application:
    """Main application information."""

    def __init__(self, settings: Settings, command: Optional[str]) -> None:
        """Load up settings."""
        self.settings = settings
        self.argument_registry = ArgumentRegistry()
        register_defaults(self.argument_registry)
        self.command_registry = CommandRegistry()
        self.register_commands()
        self.input_handler = settings.input_format.get_input_handler(
            command, settings.input_file, self.command_registry,
            settings.interactive_input.history_file_path, self._prompt,
            settings.syntax_highlighting)
        self.input_processor = settings.input_format.get_input_processor()
        self.output_handler = settings.output_format.get_output_handler(settings.syntax_highlighting)
        self.epp_handler = settings.session.get_epp_handler()

    def _prompt(self) -> str:
        """Get prompt: username@host."""
        if not self.epp_handler.logged_in:
            return "> "
        return f"{self.epp_handler.username}@{self.epp_handler.hostname} > "

    def register_command_to_registry(
            self, name: str, command_class: Type[Union[Request, ServiceCommand]],
            extensions: Optional[Dict] = None) -> None:
        """Create and register a command."""
        arguments = {field.name: field.type for field in fields(command_class)} if is_dataclass(command_class) else {}
        command = Command(name, command_class, arguments, self.argument_registry, extensions)
        self.command_registry.register_command(command)

    def execute(self, command: Union[Request, ServiceCommand]) -> None:
        """Execute the prepared command."""
        if isinstance(command, ServiceCommand):
            command.handle(self)
        elif isinstance(command, Request):
            response = self.epp_handler.send_command(command)
            LOGGER.debug(response)
            self.output_handler.output(cast(Result, response))
        else:
            LOGGER.info("Command not sent to the server.")  # type: ignore[unreachable]

    def run_command(self, input: Any) -> None:
        """Run one command and stop."""
        LOGGER.debug("Parsing raw input...")
        command_name, args = self.input_processor(input)
        LOGGER.debug("Parsing command...")
        command = self.command_registry.get_command(command_name)
        LOGGER.debug("Parsing arguments...")
        command.arg_parser.parse_arguments(args)
        arguments = command.arg_parser.get_parsed_values()
        LOGGER.debug(f"Creating {command_name} command...")
        epp_command = command.prepare_command(arguments)
        LOGGER.debug("Executing command...")
        self.execute(epp_command)
        LOGGER.debug("Command completed.")

    def run(self) -> None:
        """Run commands continuously."""
        for command in self.input_handler:
            try:
                self.run_command(command)
            except Exception as e:
                LOGGER.error(e)

        self.epp_handler.exit()

    def register_commands(self) -> None:
        """Return a list of all commands."""
        self.register_command_to_registry("exit", ServiceExit)

        # session
        self.register_command_to_registry("login", ServiceLogin)
        self.register_command_to_registry("logout", ServiceLogout)
        self.register_command_to_registry("set-session", ServiceSwitch)

        # info
        self.register_command_to_registry("info-contact", InfoContact)
        self.register_command_to_registry("info-domain", InfoDomain)
        self.register_command_to_registry("info-keyset", InfoKeyset)
        self.register_command_to_registry("info-nsset", InfoNsset)

        # transfer
        self.register_command_to_registry("transfer-contact", TransferContact)
        self.register_command_to_registry("transfer-domain", TransferDomain)
        self.register_command_to_registry("transfer-keyset", TransferKeyset)
        self.register_command_to_registry("transfer-nsset", TransferNsset)

        # update
        self.register_command_to_registry("update-contact", UpdateContact,
                                          {"mailing": Optional[UpdateContactMailingAddressExtension]})
        self.register_command_to_registry("update-domain", UpdateDomain,
                                          {"enum": Optional[UpdateDomainEnumExtension]})
        self.register_command_to_registry("update-keyset", UpdateKeyset)
        self.register_command_to_registry("update-nsset", UpdateNsset)

        # create
        self.register_command_to_registry("create-contact", CreateContact,
                                          {"mailing": Optional[CreateContactMailingAddressExtension]})
        self.register_command_to_registry("create-domain", CreateDomain,
                                          {"enum": Optional[CreateDomainEnumExtension]})
        self.register_command_to_registry("create-keyset", CreateKeyset)
        self.register_command_to_registry("create-nsset", CreateNsset)

        # renew
        self.register_command_to_registry("renew-domain", RenewDomain,
                                          {"enum": Optional[RenewDomainEnumExtension]})

        # extensions
        self.register_command_to_registry("credit-info-request", CreditInfoRequest)
        self.register_command_to_registry("send-auth-info-contact", SendAuthInfoContact)
        self.register_command_to_registry("send-auth-info-domain", SendAuthInfoDomain)
        self.register_command_to_registry("send-auth-info-keyset", SendAuthInfoKeyset)
        self.register_command_to_registry("send-auth-info-nsset", SendAuthInfoNsset)
        self.register_command_to_registry("test-nsset", TestNsset)

        # delete
        self.register_command_to_registry("delete-contact", DeleteContact)
        self.register_command_to_registry("delete-domain", DeleteDomain)
        self.register_command_to_registry("delete-keyset", DeleteKeyset)
        self.register_command_to_registry("delete-nsset", DeleteNsset)

        # check
        self.register_command_to_registry("check-contact", CheckContact)
        self.register_command_to_registry("check-domain", CheckDomain)
        self.register_command_to_registry("check-keyset", CheckKeyset)
        self.register_command_to_registry("check-nsset", CheckNsset)

        # list
        self.register_command_to_registry("get-results", GetResults)
        self.register_command_to_registry("list-contacts", ListContacts)
        self.register_command_to_registry("list-domains", ListDomains)
        self.register_command_to_registry("list-domains-by-contact", ListDomainsByContact)
        self.register_command_to_registry("list-domains-by-keyset", ListDomainsByKeyset)
        self.register_command_to_registry("list-domains-by-nsset", ListDomainsByNsset)
        self.register_command_to_registry("list-keysets", ListKeysets)
        self.register_command_to_registry("list-keysets-by-contact", ListKeysetsByContact)
        self.register_command_to_registry("list-nssets", ListNssets)
        self.register_command_to_registry("list-nssets-by-contact", ListNssetsByContact)
        self.register_command_to_registry("list-nssets-by-ns", ListNssetsByNs)

        # help command
        help_command = HelpCommand(self.command_registry)
        self.command_registry.register_command(help_command)
