"""Representation of a dataclass argument."""
from dataclasses import fields
from typing import Any, Dict, List, Optional

from .argument import Argument, ArgumentLabelSuggestion


class DataclassArgument(Argument):
    """Representation of a dataclass argument."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Save the basic info about the argument using base class, also initializing subarguments."""
        super().__init__(*args, **kwargs)
        self._init_subarguments()

    def _init_subarguments(self) -> None:
        """Generate a list of subarguments.

        The subarguments have the correct type, they just think they're not in a sequence.

        For now, the list will have one without a number, and one with number one.

        """
        self._subarguments: List[Argument] = []
        for field in fields(self._arg_type):
            self._subarguments.append(self._factory(field.name, field.type, True))

    def _find_subarg_match(self, label: List[str]) -> Optional[Argument]:
        """Find a matching subargument.

        If the subargument is not found, return None.

        Args:
            label: the label of the lowest-level subargument, split by levels.

        """
        label = label[1:]
        for subargument in self._subarguments:
            if subargument.recognise_label(label):
                return subargument
        return None

    def get_value(self) -> Dict[str, Any]:
        """Get value of the argument, with the correct type."""
        if not self.has_value and not self._required:
            return {self._arg_label: None}

        args = {}
        for arg in self._subarguments:
            arg_value = arg.get_value()
            for key, value in arg_value.items():
                if value is not None:
                    args[key] = value

        return {self._arg_label: self._arg_type(**args)}

    def add_value(self, value_labels: List[str], value: str) -> None:
        """Recognise and add the value.

        Args:
            value_labels: the name of the lowest-level subargument, split by levels.
                            (highest-argument level first)
            value: the value which should be parsed and saved to the subargument.

        """
        matched_subargument = self._find_subarg_match(value_labels)
        if matched_subargument is None:
            raise ValueError(f"Subargument {'.'.join(value_labels[1:])} of {self._arg_label} was not found!")

        matched_subargument.add_value(value_labels, value)

    def recognise_label(self, labels: List[str]) -> bool:
        """Find out whether the labels are yours or one of your subarguments'.

        Args:
            labels: the labels of the lowest-level subargument, split by levels.
                            (highest-argument level first)

        Returns:
            True if `labels` are my (subargument's) labels.

        """
        return len(labels) > 0 and labels[0] == self._arg_label

    def available_argument_labels(self) -> List[ArgumentLabelSuggestion]:
        """Return a list of available argument labels.

        Returns:
            a list of names of (sub)arguments that this argument is willing to
            accept.

        """
        ret: List[ArgumentLabelSuggestion] = []
        auto_optional = not self._required and not self.has_value
        for x in self._subarguments:
            for name, vals, req in x.available_argument_labels():
                if auto_optional:
                    ret.append(ArgumentLabelSuggestion(self.get_subarguments_full_label(name), vals, False))
                else:
                    ret.append(ArgumentLabelSuggestion(self.get_subarguments_full_label(name), vals, req))

        return ret

    @property
    def has_value(self) -> bool:
        """Whether or not the argument already has value.

        Useful for some more complicated argument types.

        """
        for x in self._subarguments:
            if x.has_value:
                return True

        return False
