"""An argument type to handle the EPPlib `Disclose` argument type."""
from typing import Any, Dict, List

from epplib.models.common import Disclose, DiscloseField

from eppic.commands.arguments import Argument, ArgumentLabelSuggestion


class DiscloseArgument(Argument):
    """Handle disclose argument."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Save the basic info about the argument.

        Args:
            arg_name (str): the name of the argument.
            arg_type (Type): the requested argument's value type.

        """
        super().__init__(*args, **kwargs)
        self._true = self._factory(self._arg_label, str, False)
        self._false = self._factory("un" + self._arg_label, str, False)

    def add_value(self, value_labels: List[str], value: str) -> None:
        """Recognise and add the value.

        Args:
            value_labels: the name of the lowest-level subargument, split by levels.
                            (highest-argument level first)
            value: the value which should be parsed and saved to the subargument.

        """
        if self._false.recognise_label(value_labels):
            self._false.add_value(value_labels, value)

        elif self._true.recognise_label(value_labels):
            self._true.add_value(value_labels, value)

        else:
            raise ValueError("Label not recognised.")

        if self._true.has_value and self._false.has_value:
            raise ValueError("Can only use either disclose or undisclose, not both.")

    def get_value(self) -> Dict[str, Any]:
        """Get value of the argument, with the correct type."""
        if self._true.has_value and self._false.has_value:
            raise ValueError("Can only use either disclose or undisclose, not both.")

        elif not (self._true.has_value or self._false.has_value):
            if self._required:
                raise ValueError("Missing disclose argument!")
            return {self._arg_label: None}

        true_values = list(self._true.get_value().values())[0]      # this is either None or a string
        false_values = list(self._false.get_value().values())[0]    # same here

        values = true_values if self._true.has_value else false_values
        value_type = self._true.has_value

        return {self._arg_label: Disclose(value_type, set([DiscloseField(x) for x in values.split(",")]))}

    def recognise_label(self, labels: List[str]) -> bool:
        """Find out whether the labels are yours or one of your subarguments'.

        Args:
            labels: the labels of the lowest-level subargument, split by levels.
                            (highest-argument level first)

        Returns:
            True if `labels` are my (subargument's) labels.

        """
        return self._false.recognise_label(labels) or self._true.recognise_label(labels)

    def available_argument_labels(self) -> List[ArgumentLabelSuggestion]:
        """Return a list of available argument labels.

        Returns:
            a list of names of (sub)arguments that this argument is willing to
            accept.

        """
        if self.has_value:
            return []

        ret = self._false.available_argument_labels() + self._true.available_argument_labels()
        return [ArgumentLabelSuggestion(label, vals, self._required) for label, vals, _ in ret]

    @property
    def has_value(self) -> bool:
        """Whether or not the argument already has value.

        Useful for some more complicated argument types.

        """
        return self._true.has_value or self._false.has_value
