"""Handle enums."""
from typing import Any, List

from .argument import Argument, ArgumentLabelSuggestion


class EnumArgument(Argument):
    """Handle ENUM arguments."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Init base class and get enum values."""
        super().__init__(*args, **kwargs)
        self._values = [val.name for val in self._arg_type]

    def add_value(self, value_labels: List[str], value: str) -> None:
        """Recognise and add the value.

        Args:
            value_labels: the name of the lowest-level subargument, split by levels.
                            (highest-argument level first)
            value: the value which should be parsed and saved to the subargument.

        """
        if self.has_value:
            raise ValueError(f"Value for {self._arg_label} is already provided!")

        if value in self._values:
            self._value = self._arg_type[value]

        else:
            self._invalid = True
            raise TypeError(f"Supplied value to arugment {self._arg_label} is not among accepted values: " +
                            ",".join(self._values))

    def available_argument_labels(self) -> List[ArgumentLabelSuggestion]:
        """Return a list of available argument labels.

        Returns:
            a list of names of (sub)arguments that this argument is willing to
            accept.

        """
        if self._value is not None:
            return []

        return [ArgumentLabelSuggestion([self._arg_label], self._values, self._required)]
