"""Argument handling sequence objects, such as lists or sets."""
from collections.abc import Sequence as ABCSequence
from typing import Any, Dict, List, Sequence, get_args, get_origin

from .argument import Argument, ArgumentLabelSuggestion


class SequenceArgument(Argument):
    """Argument class handling sequence objects, such as lists or sets."""

    SEQUENCE_CONVERSIONS = {
        Sequence: list,
        ABCSequence: list,
        List: list,
        list: list,
        set: set,
    }

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Save the basic info about the argument using base class, also initializing subarguments."""
        super().__init__(*args, **kwargs)
        self._init_subarguments()

    def _init_subarguments(self) -> None:
        """Generate a list of subarguments.

        The subarguments have the correct type, they just think they're not in a sequence.

        For now, the list will have one without a number, and one with number one.

        """
        self._subarguments: List[Argument] = []
        self._add_subargument(self._arg_label + "_1")

    def _add_subargument(self, name: str = None) -> Argument:
        """Generate a subargument.

        The subargument's name will be this arugment's name, possibly with a
        number after it. The first (0th) argument doesn't get a number, so this
        accepts arguments even without numbering (though discouraged, possible
        ambiguity)

        """
        if name is None:
            name = self._arg_label + "_" + str(len(self._subarguments) + 1)

        inner_type = get_args(self._arg_type)[0]
        subargument = self._factory(name, inner_type, False)
        self._subarguments.append(subargument)
        return subargument

    def _find_subarg_match(self, label: List[str]) -> Argument:
        """Find a matching subargument.

        If the subargument is not found, create a new one.

        Args:
            label: the label of the lowest-level subargument, split by levels.

        """
        for subargument in self._subarguments:
            if subargument.recognise_label(label):
                return subargument
        return self._add_subargument(label[0])

    def get_value(self) -> Dict[str, Any]:
        """Get value of the argument, with the correct type."""
        transform = self.SEQUENCE_CONVERSIONS[get_origin(self._arg_type)]  # type: ignore
        ret: List[Any] = []
        for sub in self._subarguments:
            sub_val = list(sub.get_value().values())
            if sub_val != [None]:
                ret += sub_val

        ret_trans = transform(ret)
        return {self._arg_label: ret_trans}

    def add_value(self, value_labels: List[str], value: str) -> None:
        """Recognise and add the value.

        Args:
            value_labels: the name of the lowest-level subargument, split by levels.
                            (highest-argument level first)
            value: the value which should be parsed and saved to the subargument.

        """
        matched_subargument = self._find_subarg_match(value_labels)
        matched_subargument.add_value(value_labels, value)

    def recognise_label(self, labels: List[str]) -> bool:
        """Find out whether the labels are yours or one of your subarguments'.

        Args:
            labels: the labels of the lowest-level subargument, split by levels.
                            (highest-argument level first)

        Returns:
            True if `labels` are my (subargument's) labels.

        """
        return labels[0].startswith(self._arg_label)

    def available_argument_labels(self) -> List[ArgumentLabelSuggestion]:
        """Return a list of available argument labels.

        Returns:
            a list of names of (sub)arguments that this argument is willing to
            accept.

        """
        ret: List[ArgumentLabelSuggestion] = []
        for x in self._subarguments:
            for name, vals, req in x.available_argument_labels():
                ret.append(ArgumentLabelSuggestion(name, vals, False))

        if ret == []:
            self._add_subargument()
            return self.available_argument_labels()

        return ret

    @property
    def has_value(self) -> bool:
        """Whether or not the argument already has value.

        Useful for some more complicated argument types.

        """
        for x in self._subarguments:
            if x.has_value:
                return True

        return False
