"""An argument type for handling Union arguments."""
from typing import Any, Dict, List, Type, get_args

from .argument import Argument, ArgumentLabelSuggestion


class UnionArgument(Argument):
    """Argument class handling Union arguments."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Save the basic info about the argument using base class, also initializing subarguments."""
        super().__init__(*args, **kwargs)
        self._init_subarguments()

    def _init_subarguments(self) -> None:
        """Generate a list of subarguments."""
        self._subarguments: List[Argument] = []

        has_str = False
        for type in get_args(self._arg_type):
            if type == str:
                has_str = True
            else:
                self._add_subargument(type)

        if has_str:
            self._add_subargument(str)

    def _add_subargument(self, inner_type: Type) -> Argument:
        """Generate a subargument."""
        subargument = self._factory(self._arg_label, inner_type, False)
        self._subarguments.append(subargument)
        return subargument

    def get_value(self) -> Dict[str, Any]:
        """Get value of the argument, with the correct type.

        If one of the accepted values is a string, then it will be the
        "fallback" type, meaning it will be selected only if all the other
        types fail. Other than that, the priority of the types is dictated by
        their order in the [ ].

        For example, for `Union[int, enum]`, it would first try to convert the
        argument to int, and only if that failed, it would move on to try the enum.

        """
        if self._invalid:
            raise TypeError(f"Invalid value for {self._arg_label}")

        for sub in self._subarguments:
            if sub.has_value:
                return sub.get_value()

        if self._required:
            raise ValueError(f"No value inserted for {self._arg_label}.")

        return {self._arg_label: None}

    def add_value(self, value_labels: List[str], value: str) -> None:
        """Recognise and add the value.

        Args:
            value_labels: the name of the lowest-level subargument, split by levels.
                            (highest-argument level first)
            value: the value which should be parsed and saved to the subargument.

        """
        if self.has_value:
            raise ValueError(f"Value to {self._arg_label} is already present.")

        for sub in self._subarguments:
            try:
                sub.add_value(value_labels, value)
                return
            except TypeError:  # may be a different type needed.
                pass

        self._invalid = True
        raise TypeError(f"Value for {self._arg_label} is of wrong type.")

    @property
    def has_value(self) -> bool:
        """Whether or not the argument already has value.

        Useful for some more complicated argument types.

        """
        for x in self._subarguments:
            if x.has_value:
                return True

        return False

    def available_argument_labels(self) -> List[ArgumentLabelSuggestion]:
        """Return a list of available argument labels.

        Returns:
            a list of names of (sub)arguments that this argument is willing to
            accept.

        """
        if self.has_value:
            return []

        ret: Dict[str, ArgumentLabelSuggestion] = {}
        for x in self._subarguments:
            for labels, vals, req in x.available_argument_labels():
                key = ".".join(labels)
                if key not in ret.keys():
                    ret[key] = ArgumentLabelSuggestion(labels, vals, req)
                else:
                    values = ret[key].values or []
                    values += vals or []
                    ret[key] = ArgumentLabelSuggestion(labels, values or None, req)

        return list(ret.values())
