"""Command class responsible for recognising user input and arguments."""
import logging
from typing import Any, Dict, List, Type, Union

from epplib.commands.base import Command as CommandClass, Request

from .argument_parser import ArgumentParser
from .argument_registry import ArgumentRegistry
from .service import ServiceCommand

LOGGER = logging.getLogger(__name__)


class Command:
    """Command class."""

    def __init__(
        self,
        name: str,
        command_class: Type[Union[Request, ServiceCommand]],
        arguments: Dict[str, Type[Any]],
        registry: ArgumentRegistry,
        extensions: Dict[str, Type[Any]] = None
    ) -> None:
        """Initialize the command from preprocessed parameters.

        Args:
            name: the name of the command.
            command_class: the class that is returned filled with arguments when the command is called.
            arguments: a dictionary of argument names and types.
            registry: an ArgumentRegistry instance.
            extensions: a dictionary of command extensions names and types.

        """
        extensions = extensions or {}
        self.name = name
        self._command_class = command_class
        self.arg_parser = ArgumentParser({**arguments, **extensions}, registry)
        self._extensions_list = list(extensions.keys())

    def _pop_extensions(self, arguments: Dict[str, Any]) -> List[Any]:
        """Pop the extensions from arguments and return them.

        Args:
            arguments: arguments and extensions.
                This is an output argument: the extensions will be removed.

        Returns: the extracted extensions.

        """
        ret = []
        for key in list(arguments.keys()):
            if key in self._extensions_list:
                extension = arguments.pop(key)
                if extension is not None:
                    ret.append(extension)
        return ret

    def prepare_command(self, arguments: Dict[str, Any]) -> Union[Request, ServiceCommand]:
        """Build the Command Class and return it.

        Args:
            arguments: already parsed arguments of the command.

        """
        extensions = self._pop_extensions(arguments)
        command = self._command_class(**arguments)

        if isinstance(command, CommandClass):
            for extension in extensions:
                LOGGER.debug(f"Extension: {extension}")
                command.add_extension(extension)

        return command
