"""Provide help for users."""
import logging
from pathlib import Path
from typing import TYPE_CHECKING, Dict, Iterable, List, Optional, cast

import yaml

from eppic.input.input_processor import ParsedCommandArgument

from .argument_parser import ArgumentParser
from .arguments.argument import ArgumentLabelSuggestion
from .command import Command
from .command_registry import CommandRegistry
from .service import ServiceCommand

if TYPE_CHECKING:
    from ..app import Application

LOGGER = logging.getLogger(__name__)
HELP_PATH = Path(__file__).parent.parent / "help.yaml"


class HelpArgumentParser(ArgumentParser):
    """Parse arguments to the help command."""

    def __init__(self, registry: CommandRegistry) -> None:
        """Save arguments and initialize arguments list.

        Args:
            arguments: a dict of arguments with their types.
            registry: an CommandRegistry instance containing all the registered commands.

        """
        self._command_registry = registry
        self._last_command: Optional[Command] = None
        self._parsed: Dict[str, List[ArgumentLabelSuggestion]] = {}

    def parse_arguments(self, arguments: Iterable[ParsedCommandArgument]) -> None:
        """Parse arguments from user input (strings) to the requrested types."""
        arguments = list(arguments)

        command_name = ""
        if len(arguments):
            first_argument = arguments.pop(0)
            command_name = first_argument.value or ""

        if command_name not in self._command_registry.registered_command_names:
            self._parsed = self._get_matching_commands_possible_arguments(command_name)

        else:
            self._parsed = self._get_single_command_possible_arguments(command_name, arguments)

        if not command_name:
            self._last_command = None

    def _get_matching_commands_possible_arguments(self, command_name: str) -> Dict[str, List[ArgumentLabelSuggestion]]:
        ret = {}
        for command in self._command_registry.registered_command_names:
            if command_name in command:
                arguments = self._get_single_command_possible_arguments(command, [])
                ret[command] = arguments[command]

        if not ret.keys() and command_name != "":  # wrong command name.
            return self._get_matching_commands_possible_arguments("")

        return ret

    def _get_single_command_possible_arguments(
            self, command_name: str, command_arguments: Iterable[ParsedCommandArgument]
    ) -> Dict[str, List[ArgumentLabelSuggestion]]:
        if command_name == "help":
            return {command_name: [ArgumentLabelSuggestion(["command, arguments"], None, True)]}

        self._last_command = self._command_registry.get_command(command_name)
        arg_parser = self._last_command.arg_parser

        try:
            arg_parser.parse_arguments([x for x in command_arguments if x.value is not None])
        except (ValueError, TypeError):
            self._get_single_command_possible_arguments(command_name, [])
        return {command_name: arg_parser.get_available_labels()}

    def get_available_labels(self) -> List[ArgumentLabelSuggestion]:
        """Get available argument labels, potentially with values, and whether they're required."""
        if self._last_command is None:
            return [ArgumentLabelSuggestion(["command"], self._command_registry.registered_command_names, True)]
        return self._last_command.arg_parser.get_available_labels()

    def get_parsed_values(self) -> Dict[str, List[ArgumentLabelSuggestion]]:
        """Get previously parsed values from the arguments."""
        return self._parsed


class ServiceHelp(ServiceCommand):
    """Print help."""

    help_path = HELP_PATH

    def __init__(self, **kwargs: List[ArgumentLabelSuggestion]) -> None:
        """Save commands."""
        self.commands = kwargs
        self._help = Help(self.help_path)

    def handle(self, app: "Application") -> None:
        """Print help."""
        for command, command_arguments in self.commands.items():
            print(self._help_command(command, command_arguments))

    def _help_command(self, command: str, arguments: List[ArgumentLabelSuggestion]) -> str:
        helps = [self._help.get_help(command)]

        for argument_labels, values, required in arguments:
            argument = ".".join(argument_labels)
            helps.append(self._help.get_help(command, argument, required=required))

            for value in values or []:
                helps.append(self._help.get_help(command, argument, value))

        return "\n".join(helps)


class HelpCommand(Command):
    """A representation of the help command."""

    def __init__(self, registry: CommandRegistry) -> None:
        """Init command."""
        self.name = "help"
        self._command_class = ServiceHelp
        self.arg_parser = HelpArgumentParser(registry)
        self._extensions_list = []


class Help:
    """Get and parse help from file."""

    indent_multiplier = 4
    max_level = 3

    def __init__(self, input_file: Path = HELP_PATH) -> None:
        """Load help."""
        self._help = {}

        input_file = input_file.expanduser()
        if not input_file.is_file():
            LOGGER.warning("Help file not found.")
            return

        with open(input_file) as f:
            self._help = yaml.safe_load(f)

    def get_help(self, *labels_tuple: str, required: bool = True) -> str:
        """Get formatted help."""
        labels = list(labels_tuple[:self.max_level])
        indent = len(labels) - 1
        if not labels:
            return ""

        name = labels[-1] if required else f"({labels[-1]})"
        try:
            command = self._help[labels.pop(0)]
            if not labels:
                return self._format(indent, name, command["description"])

            argument = command["arguments"][labels.pop(0)]
            if not labels:
                ret = argument if isinstance(argument, str) else cast(str, argument["description"])
                return self._format(indent, name, ret)

            value = argument["values"][labels.pop(0)]
            return self._format(indent, name, value)
        except KeyError:
            return self._format(indent, name, "")

    def _format(self, indent: int, name: str, help: str) -> str:
        """Format help."""
        return self.indent_multiplier * indent * ' ' + f"* {name}: {help}".strip()
