"""Service Commands implementation."""
import sys
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING, Union

if TYPE_CHECKING:
    from ..app import Application


class ServiceCommand(ABC):
    """Base class. Provides information to the app."""

    @abstractmethod
    def handle(self, app: "Application") -> None:
        """Handle the service command."""


class ServiceExit(ServiceCommand):
    """Allows the program to exit."""

    def handle(self, app: "Application") -> None:
        """Handle the service command."""
        app.epp_handler.exit()
        sys.exit(0)


@dataclass
class ServiceLogin(ServiceCommand):
    """Handle login."""

    username: str
    password: str

    def handle(self, app: "Application") -> None:
        """Handle the service command."""
        app.epp_handler.login(self.username, self.password)


class ServiceLogout(ServiceCommand):
    """Handle logout."""

    def handle(self, app: "Application") -> None:
        """Handle the service command."""
        app.epp_handler.logout()


@dataclass
class ServiceSwitch(ServiceCommand):
    """Switch Service instance."""

    session_id: Union[str, int]

    def handle(self, app: "Application") -> None:
        """Handle the service command."""
        app.settings.set_session(self.session_id)
        app.epp_handler = app.settings.session.get_epp_handler()
