"""EPP Handler, carries the EPP session."""
import logging
from pathlib import Path
from typing import List, Optional, cast

from epplib.client import Client
from epplib.commands import Login, Logout
from epplib.commands.base import Request
from epplib.responses import Response, Result
from epplib.transport import SocketTransport

LOGGER = logging.getLogger(__name__)


class EppHandler:
    """Wrapper around the EPPlib."""

    def __init__(
            self,
            hostname: str, port: int, cert_file: Path, key_file: Path, obj_uris: List[str],
            username: str = None, password: str = None, verify: bool = True
    ) -> None:
        self._cert_file = cert_file
        self._key_file = key_file
        self._obj_uris = obj_uris

        self.logged_in = False

        self.hostname = hostname
        self.port = port
        self._cert_file = cert_file
        self._key_file = key_file
        self.username: Optional[str] = None

        transport = SocketTransport(self.hostname, self.port,
                                    cert_file=self._cert_file, key_file=self._key_file, verify=verify)

        self._client = Client(transport)

        if username is not None and password is not None:
            LOGGER.debug("Login at init")
            self.login(username, password)

    def logout(self) -> None:
        """Send a logout command."""
        self.username = None
        self.logged_in = False
        self.send_command(Logout())

    def login(self, username: str, password: str) -> None:
        """Send a login command."""
        self.exit()
        self.username = username
        self._client.connect()
        response = cast(Result, self.send_command(Login(cl_id=username, password=password, obj_uris=self._obj_uris)))
        if response.code >= 2000:
            self.exit()
            if response.code in (2200, 2501):
                raise ValueError("Invalid username or password.")
            else:
                raise ValueError("Login was unsuccessful.")

        self.logged_in = True

    def exit(self) -> None:
        """Disconnects from the EPP server."""
        try:
            if self.logged_in:
                self.logout()

            self._client.close()
        except Exception:
            LOGGER.warning("Connection failed on closing!")

    def send_command(self, command: Request) -> Response:
        """Send an EPP command to the server."""
        LOGGER.info(f"Send command: {command}")
        response = self._client.send(command)

        return response
