"""Input Processor class."""
from abc import ABC, abstractmethod
from shlex import shlex
from typing import Any, Generator, Iterable, List, NamedTuple, Optional, Tuple


class ParsedCommandArgument(NamedTuple):
    """A pair of labels, value."""

    labels: Optional[List[str]]
    value: Optional[str]


class InputProcessor(ABC):
    """Parses raw input into the command name and args."""

    def __call__(self, raw_in: Any) -> Tuple[str, Iterable[ParsedCommandArgument]]:
        """Get the input."""
        cmd, args = self._get_command_name_and_raw_args(raw_in)
        return cmd, self._get_args_generator(args)

    @abstractmethod
    def _get_command_name_and_raw_args(self, raw_in: Any) -> Tuple[str, Any]:
        """Get command name."""

    @abstractmethod
    def _get_args_generator(self, args: Any) -> Iterable[ParsedCommandArgument]:
        """Create a generator for arguments."""


class StringInputProcessor(InputProcessor):
    """Process string input."""

    def _get_command_name_and_raw_args(self, raw_in: Any) -> Tuple[str, Any]:
        """Get command name."""
        if not isinstance(raw_in, str):
            raise TypeError("Wrong type to insert into the input processor.")

        if " " in raw_in:
            name, args = raw_in.split(" ", 1)
            return name, args
        return raw_in, ""

    def _get_args_generator(self, args: Any) -> Iterable[ParsedCommandArgument]:
        """Create a generator for arguments."""
        s = shlex(args, posix=True, punctuation_chars=True)
        s.wordchars += "@+"
        s.whitespace += "="

        next_labels: Optional[List[str]] = None
        for arg in s:
            if arg.startswith("--") and next_labels is None:
                next_labels = arg[2:].split(".") if arg != "--" else []
            else:
                yield ParsedCommandArgument(labels=next_labels, value=arg)
                next_labels = None

        if next_labels is not None:
            yield ParsedCommandArgument(labels=next_labels, value=None)


class DictInputProcessor(InputProcessor):
    """Process dict input."""

    def _get_command_name_and_raw_args(self, raw_in: Any) -> Tuple[str, Any]:
        """Get command name."""
        if not isinstance(raw_in, dict):
            raise TypeError("Wrong type to insert into the input processor.")

        if len(raw_in.keys()) != 1:
            raise ValueError("Only accepted one command per dictionary.")

        key = list(raw_in.keys())[0]
        return key, raw_in[key]

    def _get_args_generator(self, args: Any) -> Generator[ParsedCommandArgument, None, None]:
        """Create a generator for arguments."""
        for key, value in args.items():
            if isinstance(value, dict):

                # this basically flattens the inner dictionaries...
                for subkey, subvalue in self._get_args_generator(value):
                    yield ParsedCommandArgument(labels=[key] + subkey if subkey is not None else [key], value=subvalue)

            elif isinstance(value, list):

                for i, list_value in enumerate(value):
                    yield from self._get_args_generator({f"{key}_{i}": list_value})

            else:
                yield ParsedCommandArgument(labels=[key], value=value)
