from dataclasses import dataclass
from unittest import TestCase
from unittest.mock import Mock, patch

from epplib.commands.base import Command as EppCommand

from eppic.app import Application
from eppic.commands.service import ServiceExit, ServiceLogin, ServiceLogout, ServiceSwitch
from eppic.settings import Session, Settings


@dataclass
class DummyCommand(EppCommand):
    foobar: str

    def _get_command_payload(self) -> None:
        """Dummy method."""


@patch("eppic.settings.EppHandler")
class TestApplication(TestCase):

    def setUp(self) -> None:
        """Create dummy settings."""
        self.settings = Settings(sessions=[
            Session(hostname="asdf", port=123, username="asdf",
                    password="asdf", obj_uris=["asdf"], cert_file=__file__, key_file=__file__),
            Session(hostname="asdf", port=123, username="asdff",
                    password="asdf", obj_uris=["asdf"], cert_file=__file__, key_file=__file__),
        ], output_format="DEFAULT")

    def test_init(self, epp_handler):
        app = Application(self.settings, "asdf")
        self.assertGreater(len(app.argument_registry._types), 0)
        self.assertGreater(len(app.command_registry._commands), 0)
        epp_handler.assert_called_once()

    @patch("eppic.commands.service.sys")
    def test_service_exit(self, sys, epp_handler):
        Application(self.settings, "asdf").execute(ServiceExit())
        sys.exit.assert_called_once_with(0)

    def test_service_logout(self, epp_handler):
        Application(self.settings, "asdf").execute(ServiceLogout())
        epp_handler.return_value.logout.assert_called_once()

    def test_service_login(self, epp_handler):
        Application(self.settings, "asdf").execute(ServiceLogin("asdf", "asdf"))
        epp_handler.return_value.login.assert_called_once_with("asdf", "asdf")

    def test_service_switch(self, epp_handler):
        app = Application(self.settings, "asdf")
        app.execute(ServiceSwitch("asdff"))
        self.assertEqual(epp_handler.call_args.kwargs["username"], "asdff")

    def test_prompt_filled(self, epp_handler):
        epp_handler.return_value.username = "foo"
        epp_handler.return_value.hostname = "hello"
        self.assertEqual(Application(self.settings, "asdf")._prompt(), "foo@hello > ")

    def test_prompt_empty(self, epp_handler):
        epp_handler.return_value.logged_in = False
        self.assertEqual(Application(self.settings, "asdf")._prompt(), "> ")

    def test_execute_command(self, epp_handler):
        command = DummyCommand("asdf")
        Application(self.settings, "asdf").execute(command)
        epp_handler.return_value.send_command.assert_called_once_with(command)

    def test_run_command(self, epp_handler):
        app = Application(self.settings, "asdf")
        app.register_command_to_registry("foobar", DummyCommand, None)
        app.run_command("foobar asdf")
        epp_handler.return_value.send_command.assert_called_once_with(DummyCommand("asdf"))

    def test_run_command_fail(self, epp_handler):
        app = Application(self.settings, "foobar foo")
        app.register_command_to_registry("foobar", DummyCommand, None)
        epp_handler.return_value.send_command.side_effect = ValueError()
        app.run()
        epp_handler.return_value.send_command.assert_called_once_with(DummyCommand("foo"))

    def test_run_command_parsing_fail(self, epp_handler):
        app = Application(self.settings, "asdf")
        app.register_command_to_registry("foobar", DummyCommand, None)
        epp_handler.return_value.send_command.side_effect = ValueError()
        app.run()
        epp_handler.return_value.exit.assert_called_once()

    def test_run_command_none(self, epp_handler):
        app = Application(self.settings, "asdf")
        cmd_mock = Mock()
        cmd_mock.execute.return_value = None
        app.register_command_to_registry("foobar", cmd_mock, None)
        epp_handler.return_value.send_command.side_effect = ValueError()
        with patch("eppic.app.LOGGER") as mock:
            app.run_command("foobar")
            mock.info.assert_called_once()

    def test_run(self, epp_handler):
        app = Application(self.settings, "foobar asdf")
        app.register_command_to_registry("foobar", DummyCommand, None)
        app.run()
        self.assertEqual(epp_handler.return_value.send_command.call_args.args, (DummyCommand("asdf"),))
