"""Test the Argument class."""
import datetime
from dataclasses import dataclass
from enum import Enum
from typing import List, Optional, Sequence, Type, Union, get_args
from unittest import TestCase
from unittest.mock import Mock

from epplib.models.common import Disclose, DiscloseField

from eppic.commands.arguments import (Argument, ArgumentLabelSuggestion, BoolArgument, DataclassArgument, DateArgument,
                                      DateTimeArgument, DiscloseArgument, EnumArgument, SequenceArgument, UnionArgument)


class TestArgument(TestCase):
    """Test Argument."""

    @property
    def argument(self) -> Argument:
        """Generate a sample argument."""
        return Argument("foobar", int, True, Mock())

    def test_get_subarguments_full_label(self) -> None:
        """Test get_subarguments_full_label."""
        self.assertEqual(
            self.argument.get_subarguments_full_label(["baz"]),
            ["foobar", "baz"]
        )

    def test_recognise_label_true(self) -> None:
        """Test correct labels."""
        self.assertTrue(self.argument.recognise_label(["foobar"]))

    def test_recognise_label_false(self) -> None:
        """Test incorrect labels."""
        examples: List[List[str]] = [
            ["foobaar"],
            ["foobar", "asdf"],
            ["1foobar"],
            ["foobar1"],
            ["foobarasdf", "baz"],
            ["asdf", "foobar"],
            [],
        ]

        for example in examples:
            with self.subTest(example=example):
                self.assertFalse(self.argument.recognise_label(example))

    def test_value_none(self):
        argument = self.argument
        self.assertRaises(ValueError, argument.get_value)

    def test_value_invalid(self):
        argument = self.argument
        self.assertRaises(TypeError, argument.add_value, ["foobar"], "asdf")
        self.assertRaises(TypeError, argument.get_value)

    def test_value_ok(self):
        argument = self.argument
        argument.add_value(["foobar"], "42")
        self.assertDictEqual(argument.get_value(), {"foobar": 42})

    def test_value_multiple(self):
        argument = self.argument
        argument.add_value(["foobar"], "42")
        self.assertRaises(ValueError, argument.add_value, ["foobar"], "42")

    def test_available_labels_w_value(self):
        argument = self.argument
        argument.add_value(["foobar"], "42")
        self.assertEqual(argument.available_argument_labels(), [])

    def test_available_labels_wo_value(self):
        argument = self.argument
        self.assertEqual(argument.available_argument_labels(), [ArgumentLabelSuggestion(["foobar"], None, True)])


class SampleEnum(Enum):
    FOO = "foo"
    BAZ = "baz"


class TestEnumArgument(TestCase):
    """Test Argument."""

    @property
    def argument(self) -> Argument:
        """Generate a sample argument."""
        return EnumArgument("foobar", SampleEnum, True, Mock())

    def test_value_none(self):
        argument = self.argument
        self.assertRaises(ValueError, argument.get_value)

    def test_value_invalid(self):
        argument = self.argument
        self.assertRaises(TypeError, argument.add_value, ["foobar"], "BAR")
        self.assertRaises(TypeError, argument.get_value)

    def test_value_ok(self):
        argument = self.argument
        argument.add_value(["foobar"], "FOO")
        self.assertDictEqual(argument.get_value(), {"foobar": SampleEnum.FOO})

    def test_value_multiple(self):
        argument = self.argument
        argument.add_value(["foobar"], "FOO")
        self.assertRaises(ValueError, argument.add_value, ["foobar"], "BAZ")

    def test_available_labels_w_value(self):
        argument = self.argument
        argument.add_value(["foobar"], "FOO")
        self.assertEqual(argument.available_argument_labels(), [])

    def test_available_labels_wo_value(self):
        argument = self.argument
        self.assertEqual(
            argument.available_argument_labels(),
            [ArgumentLabelSuggestion(["foobar"], ["FOO", "BAZ"], True)])


class TestSequenceArgument(TestCase):
    """Test SequenceArgument class."""

    @property
    def argument(self) -> Argument:
        """Generate a sample argument."""
        return SequenceArgument("foobar", Sequence[int], True, lambda x, y, z: Argument(x, int, z, Mock()))

    def test_get_value_none(self):
        self.assertDictEqual(self.argument.get_value(), {"foobar": []})

    def test_get_value_many(self):
        argument = self.argument
        argument.add_value(["foobar_1"], "32")
        argument.add_value(["foobar_2"], "10")
        self.assertDictEqual(argument.get_value(), {"foobar": [32, 10]})

    def test_add_value_invalid(self):
        argument = self.argument
        self.assertRaises(TypeError, argument.add_value, ["foobar_1"], "adf")

    def test_add_value_ok_and_invalid(self):
        argument = self.argument
        argument.add_value(["foobar_1"], "32")
        self.assertRaises(TypeError, argument.add_value, ["foobar_2"], "adf")

    def test_available_labels_wo_value(self):
        self.assertListEqual(
            self.argument.available_argument_labels(),
            [ArgumentLabelSuggestion(["foobar_1"], None, False)])

    def test_available_labels_w_value(self):
        argument = self.argument
        argument.add_value(["foobar_1"], "32")
        self.assertListEqual(argument.available_argument_labels(), [ArgumentLabelSuggestion(["foobar_2"], None, False)])

    def test_has_value_true(self):
        argument = self.argument
        argument.add_value(["foobar_1"], "32")
        self.assertTrue(argument.has_value)

    def test_has_value_false(self):
        self.assertFalse(self.argument.has_value)

    def test_recognise_labels(self):
        tests = [
            ([""], False),
            (["foobar"], True),
            (["foobar_1"], True),
            (["foobar_213asdfasdf"], True),
            (["fobaar"], False),
            (["foobar_1", "adsf"], True),
            (["foobar__asdf"], True),
        ]
        for val, exp in tests:
            with self.subTest(val=val, exp=exp):
                self.assertEqual(self.argument.recognise_label(val), exp)


@dataclass
class SampleDataclass:
    """Sample testing dataclass."""

    foo: int
    bar: Optional[int] = None


class TestDataclassArgument(TestCase):
    """Test SequenceArgument class."""

    @property
    def argument(self) -> Argument:
        """Generate a sample argument."""
        return DataclassArgument(
            "foobar",
            SampleDataclass,
            True,
            lambda x, y, z: Argument(x, int, type(None) not in get_args(y), Mock())
        )

    def test_get_value_none(self):
        argument = self.argument
        argument._required = False
        self.assertDictEqual(argument.get_value(), {"foobar": None})

    def test_get_value(self):
        argument = self.argument
        argument.add_value(["foobar", "foo"], 42)
        self.assertEqual(argument.get_value(), {"foobar": SampleDataclass(42)})

    def test_add_value_invalid(self):
        argument = self.argument
        self.assertRaises(TypeError, argument.add_value, ["foobar", "foo"], "adf")

    def test_add_value_ok_and_invalid(self):
        argument = self.argument
        argument.add_value(["foobar", "foo"], "32")
        self.assertRaises(TypeError, argument.add_value, ["foobar", "bar"], "adf")

    def test_add_value_invalid_arg_name(self):
        argument = self.argument
        self.assertRaises(ValueError, argument.add_value, ["foobar", "asdf"], "42")

    def test_available_labels_wo_value(self):
        self.assertListEqual(
            self.argument.available_argument_labels(),
            [
                ArgumentLabelSuggestion(["foobar", "foo"], None, True),
                ArgumentLabelSuggestion(["foobar", "bar"], None, False)
            ]
        )

    def test_available_labels_auto_optional(self):
        argument = self.argument
        argument._required = False
        self.assertListEqual(
            argument.available_argument_labels(),
            [
                ArgumentLabelSuggestion(["foobar", "foo"], None, False),
                ArgumentLabelSuggestion(["foobar", "bar"], None, False)
            ]
        )

    def test_available_labels_w_value(self):
        argument = self.argument
        argument.add_value(["foobar", "foo"], "32")
        self.assertListEqual(
            argument.available_argument_labels(),
            [ArgumentLabelSuggestion(["foobar", "bar"], None, False)])

    def test_has_value_true(self):
        argument = self.argument
        argument.add_value(["foobar", "foo"], "32")
        self.assertTrue(argument.has_value)

    def test_has_value_false(self):
        self.assertFalse(self.argument.has_value)

    def test_recognise_labels(self):
        tests = [
            ([""], False),
            (["foobar"], True),
            (["foobar_1"], False),
            (["foobar_213asdfasdf"], False),
            (["fobaar"], False),
            (["foobar_1", "adsf"], False),
            (["foobar__asdf"], False),
            (["foobar", "foo"], True),
            (["foobar", "bar"], True),
        ]
        for val, exp in tests:
            with self.subTest(val=val, exp=exp):
                self.assertEqual(self.argument.recognise_label(val), exp)


class TestDiscloseArgument(TestCase):
    """Test Argument."""

    @property
    def argument(self) -> DiscloseArgument:
        """Generate a sample argument."""
        return DiscloseArgument("foobar", Disclose, True, lambda x, y, z: Argument(x, str, z, Mock()))

    def expected(self, disclose: bool, values: List[str]) -> Disclose:
        return Disclose(
            disclose,
            set([DiscloseField(x) for x in values])
        )

    def test_recognition(self):
        argument = self.argument
        labels = ["foobar", "unfoobar"]
        for label in labels:
            with self.subTest(label=label):
                self.assertTrue(argument.recognise_label([label]))

    def test_value_none(self):
        argument = self.argument
        self.assertRaises(ValueError, argument.get_value)

    def test_value_ok_multiple(self):
        argument = self.argument
        argument.add_value(["foobar"], "email,addr")
        self.assertDictEqual(argument.get_value(), {"foobar": self.expected(True, ["email", "addr"])})

    def test_value_ok_disclose(self):
        argument = self.argument
        argument.add_value(["foobar"], "email")
        self.assertDictEqual(argument.get_value(), {"foobar": self.expected(True, ["email"])})

    def test_value_ok_undisclose(self):
        argument = self.argument
        argument.add_value(["unfoobar"], "email")
        self.assertDictEqual(argument.get_value(), {"foobar": self.expected(False, ["email"])})

    def test_value_multiple(self):
        argument = self.argument
        argument.add_value(["foobar"], "email")
        self.assertRaises(ValueError, argument.add_value, ["foobar"], "BAZ")

    def test_value_unknown_label(self):
        argument = self.argument
        self.assertRaises(ValueError, argument.add_value, ["asdffoobar"], "addr")

    def test_value_both(self):
        argument = self.argument
        argument.add_value(["foobar"], "email")
        self.assertRaises(ValueError, argument.add_value, ["unfoobar"], "addr")
        self.assertRaises(ValueError, argument.get_value)

    def test_available_labels_w_value(self):
        argument = self.argument
        argument.add_value(["foobar"], "email")
        self.assertEqual(argument.available_argument_labels(), [])

    def test_optional(self):
        argument = self.argument
        argument._required = False
        self.assertEqual(argument.get_value(), {"foobar": None})

    def test_available_labels_wo_value(self):
        argument = self.argument
        self.assertEqual(argument.available_argument_labels(), [
            ArgumentLabelSuggestion(["unfoobar"], None, True),
            ArgumentLabelSuggestion(["foobar"], None, True)
        ])


class TestDateTimeArgument(TestCase):
    """Test DateTime class."""

    @property
    def argument(self) -> Argument:
        """Generate a sample argument."""
        return DateTimeArgument(
            "foobar",
            datetime.datetime,
            True,
            Mock()
        )

    def test_get_value(self):
        argument = self.argument
        argument.add_value(["foobar"], "2022-09-09T07:58:20")
        self.assertEqual(argument.get_value(),
                         {"foobar": datetime.datetime(year=2022, month=9, day=9, hour=7, minute=58, second=20)})

    def test_wrong_format(self):
        argument = self.argument
        self.assertRaises(TypeError, argument.add_value, ["foobar"], "2022-sdfa09-09T07:58:20")
        self.assertRaises(TypeError, argument.get_value)

    def test_value_multiple(self):
        argument = self.argument
        argument.add_value(["foobar"], "2022-09-09T07:58:20")
        self.assertRaises(ValueError, argument.add_value, ["foobar"], "2022-09-09T07:58:20")


class TestDateArgument(TestCase):
    """Test DateTime class."""

    @property
    def argument(self) -> Argument:
        """Generate a sample argument."""
        return DateArgument(
            "foobar",
            datetime.date,
            True,
            Mock()
        )

    def test_get_value(self):
        argument = self.argument
        argument.add_value(["foobar"], "2022-09-09")
        self.assertEqual(argument.get_value(),
                         {"foobar": datetime.date(year=2022, month=9, day=9)})

    def test_value_multiple(self):
        argument = self.argument
        argument.add_value(["foobar"], "2022-09-09")
        self.assertRaises(ValueError, argument.add_value, ["foobar"], "2022-09-08")

    def test_wrong_format(self):
        argument = self.argument
        self.assertRaises(TypeError, argument.add_value, ["foobar"], "2022-sdfa09-09T07:58:20")
        self.assertRaises(TypeError, argument.get_value)


class TestBoolArgument(TestCase):
    """Test Argument."""

    @property
    def argument(self) -> Argument:
        """Generate a sample argument."""
        return BoolArgument("foobar", bool, True, Mock())

    def test_value_invalid(self):
        argument = self.argument
        self.assertRaises(TypeError, argument.add_value, ["foobar"], "BAR")
        self.assertRaises(TypeError, argument.get_value)

    def test_value_ok_true(self):
        accepted_true = ["y", "Y", "t", "T"]
        for val in accepted_true:
            with self.subTest(value=val):
                argument = self.argument
                argument.add_value(["foobar"], val)
                self.assertDictEqual(argument.get_value(), {"foobar": True})

    def test_value_ok_false(self):
        accepted_true = ["f", "F", "n", "N"]
        for val in accepted_true:
            with self.subTest(value=val):
                argument = self.argument
                argument.add_value(["foobar"], val)
                self.assertDictEqual(argument.get_value(), {"foobar": False})

    def test_value_multiple(self):
        argument = self.argument
        argument.add_value(["foobar"], "y")
        self.assertRaises(ValueError, argument.add_value, ["foobar"], "y")


class TestUnionArgument(TestCase):
    """Test Union argument."""

    @property
    def argument(self) -> Argument:
        """Generate a sample argument."""
        return UnionArgument(
            "foobar",
            Union[str, int],
            True,
            self._factory
        )

    def _factory(self, name: str, type: Type, required: Optional[bool]) -> Argument:
        if type == SampleDataclass:
            return DataclassArgument(name, type, required, self._factory)

        return Argument(name, type, required or False, self._factory)

    def test_value_none(self):
        argument = self.argument
        self.assertRaises(ValueError, argument.get_value)

    def test_value_no_str(self):
        argument = self.argument
        argument.add_value(["foobar"], "42")
        self.assertDictEqual(argument.get_value(), {"foobar": 42})

    def test_value_ok(self):
        argument = self.argument
        argument.add_value(["foobar"], "asdf")
        self.assertDictEqual(argument.get_value(), {"foobar": "asdf"})

    def test_value_multiple(self):
        argument = self.argument
        argument.add_value(["foobar"], "42")
        self.assertRaises(ValueError, argument.add_value, ["foobar"], "42")

    def test_available_labels_w_value(self):
        argument = self.argument
        argument.add_value(["foobar"], "42")
        self.assertEqual(argument.available_argument_labels(), [])

    def test_available_labels_different_subarguments(self):
        argument = UnionArgument("foobar", Union[int, SampleDataclass], True, self._factory)
        self.assertEqual(argument.available_argument_labels(), [
            ArgumentLabelSuggestion(["foobar"], None, False),
            ArgumentLabelSuggestion(["foobar", "foo"], None, False),
            ArgumentLabelSuggestion(["foobar", "bar"], None, False),
        ])

    def test_available_labels_different_subarguments_w_value(self):
        argument = self.argument
        argument.add_value(["foobar"], "42")
        self.assertEqual(argument.available_argument_labels(), [])

    def test_available_labels_wo_value(self):
        argument = self.argument
        self.assertEqual(argument.available_argument_labels(), [ArgumentLabelSuggestion(["foobar"], None, False)])

    def test_value_wrong_type(self):
        argument = UnionArgument(
            "foobar",
            Union[int, float],
            True,
            lambda x, y, z: Argument(x, y, True, Mock())
        )
        self.assertRaises(TypeError, argument.add_value, ["foobar"], "asdf")
        self.assertRaises(TypeError, argument.get_value)

    def test_get_value_none(self):
        argument = self.argument
        argument._required = False
        self.assertDictEqual(argument.get_value(), {"foobar": None})
