import tempfile
from pathlib import Path
from unittest import TestCase
from unittest.mock import Mock, patch

from epplib.commands.base import Login, Logout

from eppic.epp import EppHandler
from eppic.settings import Session


@patch("eppic.epp.Client")
@patch("eppic.epp.SocketTransport")
class TestEppHandler(TestCase):
    def setUp(self) -> None:
        self.temp_dir = tempfile.TemporaryDirectory()
        keys = Path(self.temp_dir.name)
        with open(keys / "key", "w") as f:
            f.write("foobar")
        with open(keys / "cert", "w") as f:
            f.write("foobar")

        self.init = {
            "hostname": "hello",
            "port": 123,
            "cert_file": keys / "cert",
            "key_file": keys / "key",
            "username": "foo",
            "password": "bar",
            "obj_uris": ["foo", "bar"]
        }

    def tearDown(self) -> None:
        self.temp_dir.cleanup()

    def test_login(self, socket_transport, client):
        self.init.pop("username")
        self.init.pop("password")
        client.return_value.send.return_value.code = 1000
        epp = EppHandler(**self.init)
        epp.login(username="asdf", password="asdff")
        self.assertEqual(epp.username, "asdf")
        self.assertTrue(epp.logged_in)
        self.assertIsInstance(epp._client.send.call_args_list[-1].args[0], Login)

    def test_login_invalid_connection(self, socket_transport, client):
        self.init.pop("username")
        self.init.pop("password")
        client.return_value.send.side_effect = Exception
        epp = EppHandler(**self.init)
        self.assertRaises(Exception, epp.login, username="asdf", password="asdff")
        self.assertFalse(epp.logged_in)
        self.assertIsInstance(epp._client.send.call_args_list[-1].args[0], Login)

    def test_login_invalid_unknown_err(self, socket_transport, client):
        self.init.pop("username")
        self.init.pop("password")
        client.return_value.send.return_value.code = 2345
        epp = EppHandler(**self.init)
        self.assertRaisesRegex(ValueError, "Login was unsuccessful.", epp.login, username="asdf", password="asdff")
        self.assertFalse(epp.logged_in)
        self.assertIsInstance(epp._client.send.call_args_list[-1].args[0], Login)

    def test_login_invalid_auth(self, socket_transport, client):
        self.init.pop("username")
        self.init.pop("password")
        client.return_value.send.return_value.code = 2501
        epp = EppHandler(**self.init)
        self.assertRaisesRegex(ValueError, "Invalid username or password", epp.login, username="asdf", password="asdff")
        self.assertFalse(epp.logged_in)
        self.assertIsInstance(epp._client.send.call_args_list[-1].args[0], Login)

    def test_logout(self, socket_transport, client):
        client.return_value.send.return_value.code = 1000
        epp = EppHandler(**self.init)
        epp.logout()
        self.assertEqual(epp.username, None)
        self.assertIsInstance(epp._client.send.call_args_list[-1].args[0], Logout)

    def test_closing(self, socket_transport, client):
        client.return_value.send.return_value.code = 1000
        epp = EppHandler(**self.init)
        epp._client.close = Mock()
        epp.exit()
        self.assertIsInstance(epp._client.send.call_args_list[-1].args[0], Logout)
        epp._client.close.assert_called_once_with()

    def test_closing_not_logged(self, socket_transport, client):
        self.init.pop("username")
        self.init.pop("password")
        client.return_value.send.return_value.code = 1000
        epp = EppHandler(**self.init)
        epp.exit()
        self.assertFalse(len(epp._client.send.call_args_list))
        epp._client.close.assert_called_once_with()

    def test_closing_error(self, socket_transport, client):
        client.return_value.send.return_value.code = 1000
        epp = EppHandler(**self.init)
        client.return_value.close.side_effect = ValueError("foo")
        epp.exit()  # exception is not thrown.

    def test_epp_from_factory(self, socket_transport, client):
        client.return_value.send.return_value.code = 1000
        sess = Session(**self.init)
        self.assertDictEqual(EppHandler(**self.init).__dict__, sess.get_epp_handler().__dict__)
