"""Test InputHandler classes."""
import tempfile
from json import JSONDecodeError
from typing import Any, Dict, List, Optional, Type, cast
from unittest import TestCase
from unittest.mock import Mock, patch

from prompt_toolkit.application import create_app_session
from prompt_toolkit.input import create_pipe_input
from yaml import YAMLError

from eppic.commands.argument_registry import ArgumentRegistry
from eppic.commands.arguments.argument import Argument
from eppic.commands.command import Command
from eppic.commands.command_registry import CommandRegistry
from eppic.input.input_handler import (CliArgsInputHandler, FileInputHandler, JsonInputHandler,
                                       PromptToolkitInputHandler, YamlInputHandler)
from eppic.input.input_processor import StringInputProcessor
from eppic.input.prompt_toolkit_handlers import Completer, Validator


class CliArgsInputHandlerTest(TestCase):
    """Test InputHandler from args."""

    def setUp(self) -> None:
        self.ig = CliArgsInputHandler("foo bar")

    def test_iterable(self) -> None:
        iter(self.ig)

    def test_one_output(self):
        self.assertEqual(list(self.ig), ["foo bar"])

    def test_none_output(self):
        self.ig = CliArgsInputHandler(None)
        self.assertEqual(list(self.ig), [])


class FileInputHandlerTest(TestCase):
    """Test InputHandler from file."""

    def get_with_file_contents(self, file_conts: str) -> List[str]:
        tmp = tempfile.NamedTemporaryFile()
        with open(tmp.name, "w") as f:
            f.write(file_conts)

        with open(tmp.name, "r") as f:
            return list(FileInputHandler(f))

    def test_normal(self):
        self.assertEqual(
            self.get_with_file_contents(
                "asdf asdf\nfoo bar baz\nhello"
            ),
            ["asdf asdf", "foo bar baz", "hello"]
        )

    def test_blank_lines(self):
        self.assertEqual(
            self.get_with_file_contents(
                "asdf asdf\n\nfoo bar baz\nhello"
            ),
            ["asdf asdf", "foo bar baz", "hello"]
        )

    def test_blank_lines_ending(self):
        self.assertEqual(
            self.get_with_file_contents(
                "asdf asdf\n\nfoo bar baz\nhello\n"
            ),
            ["asdf asdf", "foo bar baz", "hello"]
        )

    def test_blank_lines_ending_multiple(self):
        self.assertEqual(
            self.get_with_file_contents(
                "asdf asdf\n\n\nfoo bar baz\nhello\n\n\n\n"
            ),
            ["asdf asdf", "foo bar baz", "hello"]
        )


class JsonInputHandlerTest(TestCase):
    """Test InputHandler from json file."""

    def get_with_file_contents(self, file_conts: str) -> List[Dict]:
        tmp = tempfile.NamedTemporaryFile()
        with open(tmp.name, "w") as f:
            f.write(file_conts)

        with open(tmp.name, "r") as f:
            return list(JsonInputHandler(f))

    def test_normal(self):
        self.assertListEqual(
            self.get_with_file_contents('{"asdf": "asdf"}\n{"foo":"bar"}'),
            [{"asdf": "asdf"}, {"foo": "bar"}]
        )

    def test_invalid_format(self):
        self.assertRaises(
            JSONDecodeError,
            self.get_with_file_contents, '{\n"asdf": "asdf"}\n{"foo":"bar"}'
        )


class YamlInputHandlerTest(TestCase):
    """Test InputHandler from yaml file."""

    def get_with_file_contents(self, file_conts: str) -> List[Dict]:
        tmp = tempfile.NamedTemporaryFile()
        with open(tmp.name, "w") as f:
            f.write(file_conts)

        with open(tmp.name, "r") as f:
            return list(YamlInputHandler(f))

    def test_normal(self):
        self.assertListEqual(
            self.get_with_file_contents('''---
asdf: asdf
...
---
foo: bar
...'''),
            [{"asdf": "asdf"}, {"foo": "bar"}]
        )

    def test_invalid_format(self):
        self.assertRaises(
            YAMLError,
            self.get_with_file_contents, '''---
asdf: asdf
foo: bar
.-ůsd'''
        )


class PromptToolkitTest(TestCase):
    """Test InputHandler for prompt_toolkit."""

    commands = [
        ("foobar", {"foo": Optional[str], "bar": int}),
        ("exit", {}),
    ]

    def setUp(self) -> None:
        self.input_handler = PromptToolkitInputHandler(Mock(), Mock())

    def prompt_toolkit(self) -> PromptToolkitInputHandler:
        cr = CommandRegistry()
        ar = ArgumentRegistry()
        ar.register_argument_by_type(int, Argument)
        ar.register_argument_by_type(str, Argument)

        for command, arguments in self.commands:
            cr.register_command(Command(command, Mock(), cast(Dict[str, Type[Any]], arguments), ar))

        ip = StringInputProcessor()
        completer = Completer(command_registry=cr, input_processor=ip)
        validator = Validator(command_registry=cr, input_processor=ip)
        return PromptToolkitInputHandler(completer, validator)

    def test_get_commands(self) -> None:
        with create_pipe_input() as inp:
            with create_app_session(input=inp):
                ih = iter(self.prompt_toolkit())
                inp.send_text("foobar --bar=12\n")
                inp.send_text("\n")
                inp.send_text("exit\n")

                self.assertEqual(next(ih), "foobar --bar=12")
                self.assertEqual(next(ih), "exit")

    def test_close_eof(self) -> None:
        with create_pipe_input() as inp:
            with create_app_session(input=inp):
                ih = iter(self.prompt_toolkit())
                inp.send_text("foobar --bar=12\n")
                inp.close()

                self.assertEqual(next(ih), "foobar --bar=12")
                self.assertRaises(StopIteration, next, ih)

    @patch("eppic.input.input_handler.PromptSession")
    def test_ctrl_c(self, prompt_session: Mock) -> None:
        prompt = prompt_session.return_value.prompt
        prompt.side_effect = [KeyboardInterrupt(), EOFError()]
        ih = iter(self.prompt_toolkit())
        self.assertRaises(StopIteration, next, ih)
        self.assertEqual(len(prompt.call_args), 2)

    def test_keybind_help(self):  # as far as I know, prompt toolkit does not support better tests.
        pt = self.prompt_toolkit()
        m = Mock()
        pt.help_keybind(m)
        buf = m.current_buffer
        buf.append_to_history.assert_called_once_with()
        buf.validate_and_handle.assert_called_once_with()

        buf.transform_current_line.assert_called_once()
        self.assertEqual(buf.transform_current_line.call_args[0][0]("foobar"), "help foobar")
