"""Test InputProcessor class."""
from typing import Any, Iterator, Tuple
from unittest import TestCase

from eppic.input.input_processor import DictInputProcessor, InputProcessor, ParsedCommandArgument, StringInputProcessor


class DummyInputProcessor(InputProcessor):
    """Dummy InputProcessor class."""

    def _get_command_name_and_raw_args(self, raw_in: Any) -> Tuple[str, Any]:
        """Get command name."""
        if raw_in == "baz":
            return "foobar", "args raw"
        return "", ""  # pragma: no cover

    def _get_args_generator(self, args: Any) -> Iterator[ParsedCommandArgument]:
        """Create a generator for arguments."""
        if args == "args raw":  # pragma: no cover
            yield ParsedCommandArgument(None, "args")
            yield ParsedCommandArgument(["args"], "parsed")


class TestGenericInputProcessor(TestCase):
    """Test abstract InputProcessor on a dummy class."""

    def setUp(self) -> None:
        """Set up test case."""
        self.ip = DummyInputProcessor()

    def test_ip_call(self):
        """Test that InputProcessor calls its subfunctions properly."""
        cmd, args = self.ip("baz")
        self.assertEqual(cmd, "foobar")
        self.assertEqual(list(args), [(None, "args"), (["args"], "parsed")])


class TestStringInputProcessor(TestCase):
    """Test the string processor."""

    def test_no_names(self):
        ip = StringInputProcessor()
        cmd, args = ip("hello foo bar baz")
        self.assertEqual(cmd, "hello")
        self.assertListEqual(list(args), [(None, "foo"), (None, "bar"), (None, "baz")])

    def test_all_names(self):
        ip = StringInputProcessor()
        cmd, args = ip("hello --asdf.asdf=foo --ass=bar --ddf=baz")
        self.assertEqual(cmd, "hello")
        self.assertListEqual(list(args), [(["asdf", "asdf"], "foo"), (["ass"], "bar"), (["ddf"], "baz")])

    def test_some_names(self):
        ip = StringInputProcessor()
        cmd, args = ip("hello --asdf.asdf=foo bar --ddf=baz")
        self.assertEqual(cmd, "hello")
        self.assertListEqual(list(args), [(["asdf", "asdf"], "foo"), (None, "bar"), (["ddf"], "baz")])

    def test_hanging_name(self):
        ip = StringInputProcessor()
        cmd, args = ip("hello foo bar baz --asdf=")
        self.assertEqual(cmd, "hello")
        self.assertListEqual(list(args), [(None, "foo"), (None, "bar"), (None, "baz"), (["asdf"], None)])

    def test_no_arguments(self):
        ip = StringInputProcessor()
        cmd, args = ip("hello")
        self.assertEqual(cmd, "hello")
        self.assertListEqual(list(args), [])

    def test_shlex_quoting(self):
        ip = StringInputProcessor()
        cmd, args = ip("hello \"foo bar\" 'baz'")
        self.assertEqual(cmd, "hello")
        self.assertListEqual(list(args), [(None, "foo bar"), (None, "baz")])

    def test_wrong_type(self):
        ip = StringInputProcessor()
        self.assertRaises(
            TypeError,
            ip, {"hello": "there"}
        )


class TestDictInputProcessor(TestCase):
    """Test the dict processor."""

    def test_all_names(self):
        ip = DictInputProcessor()
        cmd, args = ip({"hello": {"asdf": {"asdf": "foo"}, "ass": "bar", "ddf": "baz"}})
        self.assertEqual(cmd, "hello")
        self.assertListEqual(list(args), [(["asdf", "asdf"], "foo"), (["ass"], "bar"), (["ddf"], "baz")])

    def test_wrong_type(self):
        ip = DictInputProcessor()
        self.assertRaises(
            TypeError,
            ip, "hello there"
        )

    def test_more_commands(self):
        ip = DictInputProcessor()
        self.assertRaises(
            ValueError,
            ip, {"hello": {"": "there"}, "hey": {"the": "re"}}
        )

    def test_list_arguments(self):
        ip = DictInputProcessor()
        cmd, args = ip({"hello": {"asdf": [{"asdf": "foo"}, "bar"], "ddf": "baz"}})
        self.assertEqual(cmd, "hello")
        self.assertListEqual(list(args), [(["asdf_0", "asdf"], "foo"), (["asdf_1"], "bar"), (["ddf"], "baz")])
