import logging
import tempfile
from pathlib import Path
from unittest import TestCase
from unittest.mock import patch

import yaml
from click.testing import CliRunner

from eppic.main import main


@patch("eppic.main.Application")
@patch("eppic.main.Settings")
class TestMain(TestCase):

    def setUp(self) -> None:
        self.tmp_dir = tempfile.TemporaryDirectory()
        self.tmp_file = Path(self.tmp_dir.name) / "settings.yaml"
        with open(Path(self.tmp_dir.name) / "key", "w") as f:
            f.write("foo")

        with open(Path(self.tmp_dir.name) / "cert", "w") as f:
            f.write("foo")

        with open(self.tmp_file, "w") as f:
            yaml.safe_dump({
                "sessions": [
                    {
                        "hostname": "asdf",
                        "port": 1234,
                        "cert_file": str(Path(self.tmp_dir.name) / "cert"),
                        "key_file": str(Path(self.tmp_dir.name) / "key"),
                        "obj_uris": ["asdf"],
                        "username": "asdf",
                        "password": "ghjkl"
                    }
                ],
            }, f)

    def tearDown(self) -> None:
        self.tmp_dir.cleanup()

    def test_main_run(self, settings, session):
        settings.return_value.logging = None
        result = CliRunner().invoke(main)
        self.assertEqual(result.exit_code, 0, msg=result.exception)
        session.assert_called_once()

    def test_main_w_command(self, settings, session):
        settings.return_value.logging = None
        result = CliRunner().invoke(main, f"--config {self.tmp_file} -- asdf sadf asdf")
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(session.call_args.args[-1], "asdf sadf asdf")

    def test_main_w_input(self, settings, session):
        settings.return_value.logging = None
        result = CliRunner().invoke(main, f"--config {self.tmp_file} --input-file {self.tmp_file}")
        self.assertEqual(result.exit_code, 0)

    def test_main_w_logging_set(self, settings, session):
        logging.basicConfig(level="INFO")
        settings.return_value.logging = {"version": 1, "root": {"level": "DEBUG"}}
        result = CliRunner().invoke(main, f"--config {self.tmp_file} --input-file {self.tmp_file}")
        self.assertEqual(logging.getLogger().level, logging.DEBUG)
        self.assertEqual(result.exit_code, 0)
