import datetime
import io
from dataclasses import dataclass
from typing import List
from unittest import TestCase
from unittest.mock import patch

from eppic.output import JsonOutputHandler, OutputHandler, YamlOutputHandler


@dataclass
class DummyResData:
    foo: str
    bar: str


@dataclass
class DummyResponse:
    code: int
    msg: str
    res_data: List[DummyResData]


class TestOutputHandler(TestCase):

    @patch('sys.stdout', new_callable=io.StringIO)
    def test_output_empty(self, mock_out):
        res = DummyResponse(1000, "msg", None)
        oh = OutputHandler()
        oh.output(res)
        self.assertEqual(
            mock_out.getvalue(),
            "msg (1000)\n\n"
        )

    @patch('sys.stdout', new_callable=io.StringIO)
    @patch('eppic.output.print_formatted_text')
    def test_output_syntax_on(self, mock_format_print, mock_out):
        res_data = [
            DummyResData("asdf", "asdf"),
            DummyResData("foo", "bar")
        ]
        res = DummyResponse(1000, "msg", res_data)
        oh = OutputHandler()
        oh.output(res)
        mock_format_print.assert_called_once()

    @patch('sys.stdout', new_callable=io.StringIO)
    def test_output_multiple_res_data(self, mock_out):
        res_data = [
            DummyResData("asdf", "asdf"),
            DummyResData("foo", "bar")
        ]
        res = DummyResponse(1000, "msg", res_data)
        oh = OutputHandler(False)
        oh.output(res)
        self.assertEqual(
            mock_out.getvalue(), """msg (1000)

bar: asdf
foo: asdf
---
bar: bar
foo: foo

"""
        )

    @patch('sys.stdout', new_callable=io.StringIO)
    def test_yaml(self, mock_out):
        res_data = [
            DummyResData("asdf", "asdf"),
            DummyResData("foo", "bar")
        ]
        res = DummyResponse(1000, "msg", res_data)
        oh = YamlOutputHandler()
        oh.output(res)
        self.assertEqual(
            mock_out.getvalue(),
            """---
msg: msg
code: 1000
data:
- foo: asdf
  bar: asdf
- foo: foo
  bar: bar
...
"""
        )

    @patch('sys.stdout', new_callable=io.StringIO)
    def test_json(self, mock_out):
        res_data = [
            DummyResData("asdf", "asdf"),
            DummyResData("foo", "bar")
        ]
        res = DummyResponse(1000, "msg", res_data)
        oh = JsonOutputHandler()
        oh.output(res)
        self.assertEqual(
            mock_out.getvalue(),
            '{"msg": "msg", "code": 1000, "data": [{"foo": "asdf", "bar": "asdf"}, {"foo": "foo", "bar": "bar"}]}\n'
        )

    @patch('sys.stdout', new_callable=io.StringIO)
    def test_json_with_newline(self, mock_out):
        res_data = [
            DummyResData("asdf", "a\nsdf"),
            DummyResData("foo", "bar")
        ]
        res = DummyResponse(1000, "msg", res_data)
        oh = JsonOutputHandler()
        oh.output(res)
        self.assertEqual(
            mock_out.getvalue(),
            '{"msg": "msg", "code": 1000, "data": [{"foo": "asdf", "bar": "a\\nsdf"}, {"foo": "foo", "bar": "bar"}]}\n'
        )

    @patch('sys.stdout', new_callable=io.StringIO)
    def test_json_with_datetime(self, mock_out):
        res_data = [
            DummyResData("foo", datetime.date(day=10, month=10, year=2020))
        ]
        res = DummyResponse(1000, "msg", res_data)
        oh = JsonOutputHandler()
        oh.output(res)
        self.assertEqual(
            mock_out.getvalue(),
            '{"msg": "msg", "code": 1000, "data": [{"foo": "foo", "bar": "2020-10-10"}]}\n'
        )
