from typing import Any, Dict, List, Optional, Tuple, Type, cast
from unittest import TestCase
from unittest.mock import Mock

from prompt_toolkit.document import Document
from prompt_toolkit.validation import ValidationError

from eppic.commands.argument_registry import ArgumentRegistry
from eppic.commands.arguments.argument import Argument
from eppic.commands.command import Command
from eppic.commands.command_registry import CommandRegistry
from eppic.input.input_processor import StringInputProcessor
from eppic.input.prompt_toolkit_handlers import Completer, Validator


class TestPromptToolkitCompletionValidation(TestCase):

    commands = [
        ("foobar", {"foo": Optional[str], "bar": int}),
        ("foobrbaz", {})
    ]

    def setUp(self) -> None:
        cr = CommandRegistry()
        ar = ArgumentRegistry()
        ar.register_argument_by_type(int, Argument)
        ar.register_argument_by_type(str, Argument)

        for command, arguments in self.commands:
            cr.register_command(Command(command, Mock(), cast(Dict[str, Type[Any]], arguments), ar))

        ip = StringInputProcessor()
        self.completer = Completer(command_registry=cr, input_processor=ip)
        self.validator = Validator(command_registry=cr, input_processor=ip)

    def assertCompleter(
        self,
        input: str,
        expected_completions: List[Tuple[str, str]] = [],
        error: bool = False
    ) -> None:
        document = Document(input)
        completions = self.completer.get_completions(document, None)
        self.assertListEqual(
            [(compl.text, compl.display_text) for compl in completions],
            expected_completions, self.completer.last_word(document)
        )
        if error:
            with self.assertRaises(ValidationError):
                self.validator.validate(document)
        else:
            self.validator.validate(document)  # tests that it doesn't raise an exception.

    def test_command_partial(self):
        self.assertCompleter(
            "foob", [("foobar", "foobar"), ("foobrbaz", "foobrbaz")]
        )

    def test_command_unknown(self):
        self.assertCompleter(
            "foobasdfasdf", [], True
        )

    def test_command_known(self):
        self.assertCompleter(
            "foobar", [("foobar", "foobar")]
        )

    def test_command_known_and_space(self):
        self.assertCompleter(
            "foobar ", [("--foo=", "(--foo=)"), ("--bar=", "--bar=")]
        )

    def test_command_known_and_more(self):
        self.assertCompleter(
            "foobar --", [("--foo=", "(--foo=)"), ("--bar=", "--bar=")], True
        )

    def test_arg_first_and_space(self):
        self.assertCompleter(
            "foobar --foo=asdf ", [("--bar=", "--bar=")]
        )

    def test_arg_name_in_progress(self):
        self.assertCompleter(
            "foobar --fo", [("--foo=", "(--foo=)")], True
        )

    def test_value_in_progress(self):
        self.assertCompleter(
            "foobar --foo=asdf", []
        )

    def test_arg_name_wrong(self):
        self.assertCompleter(
            "foobar --fo ", [], True
        )

    def test_value_wrong(self):
        self.assertCompleter(
            "foobar --bar=asdf", [], True
        )

    def test_finished_command(self):
        self.assertCompleter(
            "foobar --foo foo --bar=1", []
        )

    def test_unknown_arg(self):
        self.assertCompleter(
            "foobar --foo foo --baz", [], True
        )

    def test_typing_after_finished(self):
        self.assertCompleter(
            "foobar --foo foo --bar=1 asdf", [], True
        )

    def test_suggesting_values(self):
        compl = Completer(command_registry=Mock(), input_processor=Mock())
        self.assertListEqual(
            [("--foo=", "--foo="), ("--foo=bar", "--foo=bar"), ("--foo=baz", "--foo=baz")],
            list(compl._single_argument_completions(["foo"], ["bar", "baz"], True))
        )

    def test_errors_on_suggestions(self):
        compl = Completer(command_registry=Mock(), input_processor=Mock())
        document = Document("unknowncommand --unknownargument=foobar")
        self.assertEqual(list(compl.get_completions(document, None)), [])  # exception not thrown
